@@ -1,5 +1,5 @@
 #!/usr/bin/env python
-# $Id: CheetahWrapper.py,v 1.15 2003/02/06 05:36:13 hierro Exp $
+# $Id: CheetahWrapper.py,v 1.26 2007/10/02 01:22:04 tavis_rudd Exp $
 """Cheetah command-line interface.
 
 2002-09-03 MSO: Total rewrite.
@@ -8,36 +8,24 @@
 
 Meta-Data
 ================================================================================
-Author: Tavis Rudd <tavis@damnsimple.com> and Mike Orr <iron@mso.oz.net>
-Version: $Revision: 1.15 $
+Author: Tavis Rudd <tavis@damnsimple.com> and Mike Orr <sluggoster@gmail.com>>
+Version: $Revision: 1.26 $
 Start Date: 2001/03/30
-Last Revision Date: $Date: 2003/02/06 05:36:13 $
+Last Revision Date: $Date: 2007/10/02 01:22:04 $
 """
-__author__ = "Tavis Rudd <tavis@damnsimple.com> and Mike Orr <iron@mso.oz.net>"
-__revision__ = "$Revision: 1.15 $"[11:-2]
-
-##################################################
-## DEPENDENCIES
+__author__ = "Tavis Rudd <tavis@damnsimple.com> and Mike Orr <sluggoster@gmail.com>"
+__revision__ = "$Revision: 1.26 $"[11:-2]
 
 import getopt, glob, os, pprint, re, shutil, sys
 import cPickle as pickle
 
-from _properties import Version
-from Compiler import Compiler
-from Template import Template
+from Cheetah.Version import Version
+from Cheetah.Template import Template, DEFAULT_COMPILER_SETTINGS
 from Cheetah.Utils.Misc import mkdirsWithPyInitFiles
 from Cheetah.Utils.optik import OptionParser
 
-##################################################
-## GLOBALS & CONSTANTS
-
-try:
-    True,False
-except NameError:
-    True, False = (1==1),(1==0)
-
-optionDashesRx = re.compile(  R"^-{1,2}"  )
-moduleNameRx = re.compile(  R"^[a-zA-Z_][a-zA-Z_0-9]*$"  )
+optionDashesRE = re.compile(  R"^-{1,2}"  )
+moduleNameRE = re.compile(  R"^[a-zA-Z_][a-zA-Z_0-9]*$"  )
    
 def fprintfMessage(stream, format, *args):
     if format[-1:] == '^':
@@ -50,8 +38,6 @@ def fprintfMessage(stream, format, *args):
         message = format
     stream.write(message)
 
-die = sys.exit
-
 class Error(Exception):
     pass
 
@@ -92,7 +78,7 @@ def usage(usageMessage, errorMessage="", out=sys.stderr):
         out.write('\n')
         out.write("*** USAGE ERROR ***: %s\n" % errorMessage)
         exitStatus = 1
-    die(exitStatus)
+    sys.exit(exitStatus)
              
 
 WRAPPER_TOP = """\
@@ -101,7 +87,7 @@ WRAPPER_TOP = """\
           \/    *   *     \/    CHEETAH %(Version)s Command-Line Tool
            \      |       / 
             \  ==----==  /      by Tavis Rudd <tavis@damnsimple.com>
-             \__________/       and Mike Orr <iron@mso.oz.net>
+             \__________/       and Mike Orr <sluggoster@gmail.com>
               
 """ % globals()
 
@@ -113,7 +99,8 @@ USAGE:
   cheetah fill [options] [FILES ...]        : Fill template definitions
   cheetah help                              : Print this help message
   cheetah options                           : Print options help message
-  cheetah test                              : Run Cheetah's regression tests
+  cheetah test [options]                    : Run Cheetah's regression tests
+                                            : (same as for unittest)
   cheetah version                           : Print Cheetah version number
 
 You may abbreviate the command to the first letter; e.g., 'h' == 'help'.
@@ -127,13 +114,19 @@ OPTIONS FOR "compile" AND "fill":
   --idir DIR, --odir DIR : input/output directories (default: current dir)
   --iext EXT, --oext EXT : input/output filename extensions
     (default for compile: tmpl/py,  fill: tmpl/html)
-  -R            : recurse subdirectories looking for input files
-  --debug       : print lots of diagnostic output to standard error
-  --env         : put the environment in the searchList
-  --flat        : no destination subdirectories
-  --nobackup    : don't make backups
-  --pickle FILE : unpickle FILE and put that object in the searchList
-  --stdout, -p  : output to standard output (pipe)
+  -R                 : recurse subdirectories looking for input files
+  --debug            : print lots of diagnostic output to standard error
+  --env              : put the environment in the searchList
+  --flat             : no destination subdirectories
+  --nobackup         : don't make backups
+  --pickle FILE      : unpickle FILE and put that object in the searchList
+  --stdout, -p       : output to standard output (pipe)
+  --settings         : a string representing the compiler settings to use
+                       e.g. --settings='useNameMapper=False,useFilters=False'
+                       This string is eval'd in Python so it should contain
+                       valid Python syntax.
+  --templateAPIClass : a string representing a subclass of
+                       Cheetah.Template:Template to use for compilation
 
 Run "cheetah help" for the main help screen.
 """
@@ -143,18 +136,149 @@ Run "cheetah help" for the main help screen.
 
 class CheetahWrapper:
     MAKE_BACKUPS = True
-    BACKUP_SUFFIX = "_bak"
+    BACKUP_SUFFIX = ".bak"
+    _templateClass = None
+    _compilerSettings = None    
 
     def __init__(self):
         self.progName = None
         self.command = None
         self.opts = None
-        self.files = None
+        self.pathArgs = None
         self.sourceFiles = []
         self.searchList = []
 
     ##################################################
-    ## VERBOSITY METHODS
+    ## MAIN ROUTINE
+
+    def main(self, argv=None):
+        """The main program controller."""
+
+        if argv is None:
+            argv = sys.argv
+
+        # Step 1: Determine the command and arguments.
+        try:
+            self.progName = progName = os.path.basename(argv[0])
+            self.command = command = optionDashesRE.sub("", argv[1])
+            if command == 'test':
+                self.testOpts = argv[2:]
+            else:
+                self.parseOpts(argv[2:])
+        except IndexError:
+            usage(HELP_PAGE1, "not enough command-line arguments")
+
+        # Step 2: Call the command
+        meths = (self.compile, self.fill, self.help, self.options,
+            self.test, self.version)
+        for meth in meths:
+            methName = meth.__name__
+            # Or meth.im_func.func_name
+            # Or meth.func_name (Python >= 2.1 only, sometimes works on 2.0)
+            methInitial = methName[0]
+            if command in (methName, methInitial):
+                sys.argv[0] += (" " + methName)
+                # @@MO: I don't necessarily agree sys.argv[0] should be 
+                # modified.
+                meth()
+                return
+        # If none of the commands matched.
+        usage(HELP_PAGE1, "unknown command '%s'" % command)
+
+    def parseOpts(self, args):
+        C, D, W = self.chatter, self.debug, self.warn
+        self.isCompile = isCompile = self.command[0] == 'c'
+        defaultOext = isCompile and ".py" or ".html"
+        parser = MyOptionParser()
+        pao = parser.add_option
+        pao("--idir", action="store", dest="idir", default="")
+        pao("--odir", action="store", dest="odir", default="")
+        pao("--iext", action="store", dest="iext", default=".tmpl")
+        pao("--oext", action="store", dest="oext", default=defaultOext)
+        pao("-R", action="store_true", dest="recurse", default=False)
+        pao("--stdout", "-p", action="store_true", dest="stdout", default=False)
+        pao("--debug", action="store_true", dest="debug", default=False)
+        pao("--env", action="store_true", dest="env", default=False)
+        pao("--pickle", action="store", dest="pickle", default="")
+        pao("--flat", action="store_true", dest="flat", default=False)
+        pao("--nobackup", action="store_true", dest="nobackup", default=False)
+        pao("--settings", action="store", dest="compilerSettingsString", default=None)
+        pao("--templateAPIClass", action="store", dest="templateClassName", default=None)
+
+        self.opts, self.pathArgs = opts, files = parser.parse_args(args)
+        D("""\
+cheetah compile %s
+Options are
+%s
+Files are %s""", args, pprint.pformat(vars(opts)), files)
+
+
+        #cleanup trailing path separators
+        seps = [sep for sep in [os.sep, os.altsep] if sep]
+        for attr in ['idir', 'odir']:
+            for sep in seps:
+                path = getattr(opts, attr, None)
+                if path and path.endswith(sep):
+                    path = path[:-len(sep)]
+                    setattr(opts, attr, path)
+                    break
+
+        self._fixExts()
+        if opts.env:
+            self.searchList.insert(0, os.environ)
+        if opts.pickle:
+            f = open(opts.pickle, 'rb')
+            unpickled = pickle.load(f)
+            f.close()
+            self.searchList.insert(0, unpickled)
+        opts.verbose = not opts.stdout
+
+    ##################################################
+    ## COMMAND METHODS
+
+    def compile(self):
+        self._compileOrFill()
+
+    def fill(self):
+        from Cheetah.ImportHooks import install
+        install()        
+        self._compileOrFill()
+
+    def help(self):
+        usage(HELP_PAGE1, "", sys.stdout)
+
+    def options(self):
+        usage(HELP_PAGE2, "", sys.stdout)
+
+    def test(self):
+        # @@MO: Ugly kludge.
+        TEST_WRITE_FILENAME = 'cheetah_test_file_creation_ability.tmp'
+        try:
+            f = open(TEST_WRITE_FILENAME, 'w')
+        except:
+            sys.exit("""\
+Cannot run the tests because you don't have write permission in the current
+directory.  The tests need to create temporary files.  Change to a directory
+you do have write permission to and re-run the tests.""")
+        else:
+            f.close()
+            os.remove(TEST_WRITE_FILENAME)
+        # @@MO: End ugly kludge.
+        from Cheetah.Tests import Test
+        import Cheetah.Tests.unittest_local_copy as unittest
+        del sys.argv[1:] # Prevent unittest from misinterpreting options.
+        sys.argv.extend(self.testOpts)
+        #unittest.main(testSuite=Test.testSuite)
+        #unittest.main(testSuite=Test.testSuite)
+        unittest.main(module=Test)
+        
+    def version(self):
+        print Version
+
+    # If you add a command, also add it to the 'meths' variable in main().
+    
+    ##################################################
+    ## LOGGING METHODS
 
     def chatter(self, format, *args):
         """Print a verbose message to stdout.  But don't if .opts.stdout is
@@ -177,10 +301,16 @@ class CheetahWrapper:
         """
         fprintfMessage(sys.stderr, format, *args)
 
+    def error(self, format, *args):
+        """Always print a warning message to stderr and exit with an error code.        
+        """
+        fprintfMessage(sys.stderr, format, *args)
+        sys.exit(1)
 
     ##################################################
     ## HELPER METHODS
 
+
     def _fixExts(self):
         assert self.opts.oext, "oext is empty!"
         iext, oext = self.opts.iext, self.opts.oext
@@ -190,88 +320,31 @@ class CheetahWrapper:
             self.opts.oext = "." + oext
     
 
-    def parseOpts(self, args):
+
+    def _compileOrFill(self):
         C, D, W = self.chatter, self.debug, self.warn
-        self.isCompile = isCompile = self.command[0] == 'c'
-        defaultOext = isCompile and ".py" or ".html"
-        parser = MyOptionParser()
-        pao = parser.add_option
-        pao("--idir", action="store", dest="idir", default="")
-        pao("--odir", action="store", dest="odir", default="")
-        pao("--iext", action="store", dest="iext", default=".tmpl")
-        pao("--oext", action="store", dest="oext", default=defaultOext)
-        pao("-R", action="store_true", dest="recurse", default=0)
-        pao("--stdout", "-p", action="store_true", dest="stdout", default=0)
-        pao("--debug", action="store_true", dest="debug", default=0)
-        pao("--env", action="store_true", dest="env", default=0)
-        pao("--pickle", action="store_true", dest="pickle", default=0)
-        pao("--flat", action="store_true", dest="flat", default=0)
-        pao("--nobackup", action="store_true", dest="nobackup", default=0)
-        self.opts, self.files = opts, files = parser.parse_args(args)
-        D("""\
-cheetah compile %s
-Options are
-%s
-Files are %s""", args, pprint.pformat(vars(opts)), files)
-        self._fixExts()
-        if opts.env:
-            self.searchList.append(os.environ)
-        if opts.pickle:
-            f = open(opts.pickle, 'rb')
-            unpickled = pickle.load(f)
-            f.close()
-            self.searchList.append(unpickled)
-        opts.verbose = not opts.stdout
-
-
-    def compileOrFillStdin(self):
-            if self.isCompile:
-                output = Compiler(file=sys.stdin)
-            else:
-                output = Template(file=sys.stdin)
-            output = str(output)
-            sys.stdout.write(output)
-
-
-    def compileOrFillBundle(self, b):
-        C, D, W = self.chatter, self.debug, self.warn
-        src = b.src
-        dst = b.dst
-        base = b.base
-        basename = b.basename
-        dstDir = os.path.dirname(dst)
-        what = self.isCompile and "Compiling" or "Filling"
-        C("%s %s -> %s^", what, src, dst) # No trailing newline.
-        if os.path.exists(dst) and not self.opts.nobackup:
-            bak = b.bak
-            C(" (backup %s)", bak) # On same line as previous message.
+        opts, files = self.opts, self.pathArgs
+        if files == ["-"]: 
+            self._compileOrFillStdin()
+            return
+        elif not files and opts.recurse:
+            which = opts.idir and "idir" or "current"
+            C("Drilling down recursively from %s directory.", which)
+            sourceFiles = []
+            dir = os.path.join(self.opts.idir, os.curdir)
+            os.path.walk(dir, self._expandSourceFilesWalk, sourceFiles)
+        elif not files:
+            usage(HELP_PAGE1, "Neither files nor -R specified!")
         else:
-            bak = None
-            C("")
-        if self.isCompile:
-            if not moduleNameRx.match(basename):
-                tup = basename, src
-                raise Error("""\
-%s: base name %s contains invalid characters.  It must
-be named according to the same rules as Python modules.""" % tup)
-            obj = Compiler(file=src, \
-                moduleName=basename, mainClassName=basename)
-        else:
-            obj = Template(file=src, searchList=self.searchList)
-        output = str(obj)
-        if bak:
-            shutil.copyfile(dst, bak)
-        if dstDir and not os.path.exists(dstDir):
-            if self.isCompile:
-                mkdirsWithPyInitFiles(dstDir)
-            else:
-                os.makedirs(dstDir)
-        if self.opts.stdout:
-            sys.stdout.write(output)
-        else:
-            f = open(dst, 'w')
-            f.write(output)
-            f.close()
+            sourceFiles = self._expandSourceFiles(files, opts.recurse, True)
+        sourceFiles = [os.path.normpath(x) for x in sourceFiles]
+        D("All source files found: %s", sourceFiles)
+        bundles = self._getBundles(sourceFiles)
+        D("All bundles: %s", pprint.pformat(bundles))
+        if self.opts.flat:
+            self._checkForCollisions(bundles)
+        for b in bundles:
+            self._compileOrFillBundle(b)
 
     def _checkForCollisions(self, bundles):
         """Check for multiple source paths writing to the same destination
@@ -292,15 +365,62 @@ be named according to the same rules as Python modules.""" % tup)
             if len(sources) > 1:
                 isError = True
                 sources.sort()
-                fmt = \
-"Collision: multiple source files %s map to one destination file %s"
+                fmt = "Collision: multiple source files %s map to one destination file %s"
                 W(fmt, sources, dst)
         if isError:
             what = self.isCompile and "Compilation" or "Filling"
-            die("%s aborted due to collisions" % what)
+            sys.exit("%s aborted due to collisions" % what)
                 
 
-    def getBundles(self, sourceFiles):
+    def _expandSourceFilesWalk(self, arg, dir, files):
+        """Recursion extension for .expandSourceFiles().
+           This method is a callback for os.path.walk().
+           'arg' is a list to which successful paths will be appended.
+        """
+        iext = self.opts.iext
+        for f in files:
+            path = os.path.join(dir, f)
+            if   path.endswith(iext) and os.path.isfile(path):
+                arg.append(path)
+            elif os.path.islink(path) and os.path.isdir(path):
+                os.path.walk(path, self._expandSourceFilesWalk, arg)
+            # If is directory, do nothing; 'walk' will eventually get it.
+
+
+    def _expandSourceFiles(self, files, recurse, addIextIfMissing):
+        """Calculate source paths from 'files' by applying the 
+           command-line options.
+        """
+        C, D, W = self.chatter, self.debug, self.warn
+        idir = self.opts.idir
+        iext = self.opts.iext
+        files = [] 
+        for f in self.pathArgs:
+            oldFilesLen = len(files)
+            D("Expanding %s", f)
+            path = os.path.join(idir, f)
+            pathWithExt = path + iext # May or may not be valid.
+            if os.path.isdir(path):
+                if recurse:
+                    os.path.walk(path, self._expandSourceFilesWalk, files)
+                else:
+                    raise Error("source file '%s' is a directory" % path)
+            elif os.path.isfile(path):
+                files.append(path)
+            elif (addIextIfMissing and not path.endswith(iext) and 
+                  os.path.isfile(pathWithExt)):
+                files.append(pathWithExt)
+                # Do not recurse directories discovered by iext appending.
+            elif os.path.exists(path):
+                W("Skipping source file '%s', not a plain file.", path)
+            else:
+                W("Skipping source file '%s', not found.", path)
+            if len(files) > oldFilesLen:
+                D("  ... found %s", files[oldFilesLen:])
+        return files
+
+
+    def _getBundles(self, sourceFiles):
         flat = self.opts.flat
         idir = self.opts.idir
         iext = self.opts.iext
@@ -320,151 +440,147 @@ be named according to the same rules as Python modules.""" % tup)
             if flat:
                 dst = os.path.join(odir, basename + oext)
             else:
-                dst = os.path.join(odir, base + oext)
+                dbn = basename
+                if odir and base.startswith(os.sep):
+                    odd = odir
+                    while odd != '':
+                        idx = base.find(odd)
+                        if idx == 0:
+                            dbn = base[len(odd):]
+                            if dbn[0] == '/':
+                                dbn = dbn[1:]
+                            break
+                        odd = os.path.dirname(odd)
+                        if odd == '/':
+                            break
+                    dst = os.path.join(odir, dbn + oext)
+                else:
+                    dst = os.path.join(odir, base + oext)
             bak = dst + self.BACKUP_SUFFIX
             b = Bundle(src=src, dst=dst, bak=bak, base=base, basename=basename)
             bundles.append(b)
         return bundles
 
 
-    def _expandSourceFilesWalk(self, arg, dir, files):
-        """Recursion extension for .expandSourceFiles().
-           This method is a callback for os.path.walk().
-           'arg' is a list to which successful paths will be appended.
-        """
-        iext = self.opts.iext
-        for fil in files:
-            path = os.path.join(dir, fil)
-            if   path.endswith(iext) and os.path.isfile(path):
-                arg.append(path)
-            elif os.path.islink(path) and os.path.isdir(path):
-                os.path.walk(path, self._expandSourceFilesWalk, arg)
-            # If is directory, do nothing; 'walk' will eventually get it.
-
-
-    def expandSourceFiles(self, files, recurse, addIextIfMissing):
-        """Calculate source paths from 'files' by applying the 
-           command-line options.
-        """
+    def _getTemplateClass(self):
         C, D, W = self.chatter, self.debug, self.warn
-        idir = self.opts.idir
-        iext = self.opts.iext
-        ret = [] 
-        for fil in self.files:
-            oldRetLen = len(ret)
-            D("Expanding %s", fil)
-            path = os.path.join(idir, fil)
-            pathWithExt = path + iext # May or may not be valid.
-            if os.path.isdir(path):
-                if recurse:
-                    os.path.walk(path, self._expandSourceFilesWalk, ret)
-                else:
-                    raise Error("source file '%s' is a directory" % path)
-            elif os.path.isfile(path):
-                ret.append(path)
-            elif addIextIfMissing and not path.endswith(iext) and \
-                os.path.isfile(pathWithExt):
-                ret.append(pathWithExt)
-                # Do not recurse directories discovered by iext appending.
-            elif os.path.exists(path):
-                W("Skipping source file '%s', not a plain file.", path)
-            else:
-                W("Skipping source file '%s', not found.", path)
-            if len(ret) > oldRetLen:
-                D("  ... found %s", ret[oldRetLen:])
-        return ret
+        modname = None
+        if self._templateClass:
+            return self._templateClass
 
+        modname = self.opts.templateClassName
 
-    def compileOrFill(self):
-        C, D, W = self.chatter, self.debug, self.warn
-        opts, files = self.opts, self.files
-        if files == ["-"]: 
-            self.compileOrFillStdin()
-            return
-        elif not files and opts.recurse:
-            which = opts.idir and "idir" or "current"
-            C("Drilling down recursively from %s directory.", which)
-            sourceFiles = []
-            dir = os.path.join(self.opts.idir, os.curdir)
-            os.path.walk(dir, self._expandSourceFilesWalk, sourceFiles)
-        elif not files:
-            usage(HELP_PAGE1, "Neither files nor -R specified!")
-        else:
-            sourceFiles = self.expandSourceFiles(files, opts.recurse, True)
-        sourceFiles = [os.path.normpath(x) for x in sourceFiles]
-        D("All source files found:", sourceFiles)
-        bundles = self.getBundles(sourceFiles)
-        D("All bundles: %s", pprint.pformat(bundles))
-        if self.opts.flat:
-            self._checkForCollisions(bundles)
-        for b in bundles:
-            self.compileOrFillBundle(b)
+        if not modname:
+            return Template
+        p = modname.rfind('.')
+        if ':' not in modname:
+            self.error('The value of option --templateAPIClass is invalid\n'
+                       'It must be in the form "module:class", '
+                       'e.g. "Cheetah.Template:Template"')
             
+        modname, classname = modname.split(':')
 
-    ##################################################
-    ## COMMAND METHODS
+        C('using --templateAPIClass=%s:%s'%(modname, classname))
+        
+        if p >= 0:
+            mod = getattr(__import__(modname[:p], {}, {}, [modname[p+1:]]), modname[p+1:])
+        else:
+            mod = __import__(modname, {}, {}, [])
 
-    def compile(self):
-        self.compileOrFill()
-
-    def fill(self):
-        self.compileOrFill()
-
-    def help(self):
-        usage(HELP_PAGE1, "", sys.stdout)
-
-    def options(self):
-        usage(HELP_PAGE2, "", sys.stdout)
-
-    def test(self):
-        from Cheetah.Tests import Test
-        import Cheetah.Tests.unittest_local_copy as unittest
-        del sys.argv[1:] # Prevent unittest from misinterpreting options.
-        unittest.main(testSuite=Test.testSuite)
-
-    def version(self):
-        print Version
-
-    # If you add a command, also add it to the 'meths' variable in main().
-
-    ##################################################
-    ## MAIN ROUTINE
-
-    def main(self, argv=None):
-        """The main program controller."""
-
-        if argv is None:
-            argv = sys.argv
-
-        # Step 1: Determine the command and arguments.
-        try:
-            self.progName = progName = os.path.basename(argv[0])
-            self.command = command = optionDashesRx.sub("", argv[1])
-            self.parseOpts(argv[2:])
-        except IndexError:
-            usage(HELP_PAGE1, "not enough command-line arguments")
-
-        # Step 2: Call the command
-        meths = (self.compile, self.fill, self.help, self.options,
-            self.test, self.version)
-        for meth in meths:
-            methName = meth.__name__
-            # Or meth.im_func.func_name
-            # Or meth.func_name (Python >= 2.1 only, sometimes works on 2.0)
-            methInitial = methName[0]
-            if command in (methName, methInitial):
-                sys.argv[0] += (" " + methName)
-                # @@MO: I don't necessarily agree sys.argv[0] should be 
-                # modified.
-                meth()
-                return
-        # If none of the commands matched.
-        usage(HELP_PAGE1, "unknown command '%s'" % command)
-
-    #run = main
+        klass = getattr(mod, classname, None)
+        if klass:
+            self._templateClass = klass
+            return klass
+        else:
+            self.error('**Template class specified in option --templateAPIClass not found\n'
+                       '**Falling back on Cheetah.Template:Template')
 
 
+    def _getCompilerSettings(self):
+        if self._compilerSettings:
+            return self._compilerSettings
 
+        def getkws(**kws):
+            return kws
+        if self.opts.compilerSettingsString:
+            try:
+                exec 'settings = getkws(%s)'%self.opts.compilerSettingsString
+            except:                
+                self.error("There's an error in your --settings option."
+                          "It must be valid Python syntax.\n"
+                          +"    --settings='%s'\n"%self.opts.compilerSettingsString
+                          +"  %s: %s"%sys.exc_info()[:2] 
+                          )
+
+            validKeys = DEFAULT_COMPILER_SETTINGS.keys()
+            if [k for k in settings.keys() if k not in validKeys]:
+                self.error(
+                    'The --setting "%s" is not a valid compiler setting name.'%k)
+            
+            self._compilerSettings = settings
+            return settings
+        else:
+            return {}
+
+    def _compileOrFillStdin(self):
+        TemplateClass = self._getTemplateClass()
+        compilerSettings = self._getCompilerSettings()
+        if self.isCompile:
+            pysrc = TemplateClass.compile(file=sys.stdin,
+                                          compilerSettings=compilerSettings,
+                                          returnAClass=False)
+            output = pysrc
+        else:
+            output = str(TemplateClass(file=sys.stdin, compilerSettings=compilerSettings))
+        sys.stdout.write(output)
+
+    def _compileOrFillBundle(self, b):
+        C, D, W = self.chatter, self.debug, self.warn
+        TemplateClass = self._getTemplateClass()
+        compilerSettings = self._getCompilerSettings()
+        src = b.src
+        dst = b.dst
+        base = b.base
+        basename = b.basename
+        dstDir = os.path.dirname(dst)
+        what = self.isCompile and "Compiling" or "Filling"
+        C("%s %s -> %s^", what, src, dst) # No trailing newline.
+        if os.path.exists(dst) and not self.opts.nobackup:
+            bak = b.bak
+            C(" (backup %s)", bak) # On same line as previous message.
+        else:
+            bak = None
+            C("")
+        if self.isCompile:
+            if not moduleNameRE.match(basename):
+                tup = basename, src
+                raise Error("""\
+%s: base name %s contains invalid characters.  It must
+be named according to the same rules as Python modules.""" % tup)
+            pysrc = TemplateClass.compile(file=src, returnAClass=False,
+                                          moduleName=basename,
+                                          className=basename,
+                                          compilerSettings=compilerSettings)
+            output = pysrc
+        else:
+            #output = str(TemplateClass(file=src, searchList=self.searchList))
+            tclass = TemplateClass.compile(file=src, compilerSettings=compilerSettings)
+            output = str(tclass(searchList=self.searchList))
+            
+        if bak:
+            shutil.copyfile(dst, bak)
+        if dstDir and not os.path.exists(dstDir):
+            if self.isCompile:
+                mkdirsWithPyInitFiles(dstDir)
+            else:
+                os.makedirs(dstDir)
+        if self.opts.stdout:
+            sys.stdout.write(output)
+        else:
+            f = open(dst, 'w')
+            f.write(output)
+            f.close()
+            
 
 ##################################################
 ## if run from the command line
